# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Small NumPy datasets for debugging/testing.
"""

from __future__ import print_function as _print_function

import sys as _sys

from . import boston_housing
from . import cifar10
from . import cifar100
from . import fashion_mnist
from . import imdb
from . import mnist
from . import reuters

del _print_function
